@@ -1,6 +1,6 @@
 /* Copyright 2007 by Kim Minh Kaplan
  *
- * greyfix.c version 0
+ * greyfix.c version 0.3.1
  *
  * Postfix policy daemon designed to prevent spam using the
  * greylisting method.
@@ -14,6 +14,7 @@
 #include <config.h>
 #endif
 #include <errno.h>
+#include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -41,12 +42,27 @@
  * update_record_life and update_record_life_secs.
  */
 #define AUTO_RECORD_LIFE_SECS (5 * 3600) /* 5 hours */
+/**
+ * How much life (in secs) to give to a record we are updating from an
+ * allowed (passed) email.
+ *
+ * The default is 36 days, which should be enough to handle messages
+ * that may only be sent once a month, or on things like the first
+ * monday of the month (which sometimes means 5 weeks).  Plus, we add
+ * a day for a delivery buffer.
+ */
+#define UPDATE_RECORD_LIFE_SECS (36 * 24 * 3600)
 
 #define DEF_DB_HOME DATA_STATE_DIR"/"PACKAGE
 
+#define DB_STAT_NAME "stats.db"
 #define DB_FILE_NAME "triplets.db"
 #define SEP '\000'
 
+/* Counter for triplets that are expired and never passed anything
+ * through */
+#define STAT_BLOCKED "blocked"
+
 #define prefixp(s,p) (!strncmp((s),(p),sizeof(p)-1))
 
 struct triplet_data {
@@ -67,6 +83,7 @@ static int opt_verbose = 0;
 
 static DB_ENV *dbenv = 0;
 static DB *db = 0;
+static DB *statsdb = 0;
 
 static char *policy_request = 0;
 static size_t policy_request_size = 0;
@@ -122,6 +139,19 @@ prepare_db()
 		      DB_CREATE, 0644);
 	if (rc)
 	    log_db_error("db->open", rc);
+	else {
+	    rc = db_create(&statsdb, dbenv, 0);
+	    if (rc)
+		syslog(LOG_ERR, "BDB-%d: db_create statsdb: %s",
+		       rc, db_strerror(rc));
+	    else {
+		rc = db->open(statsdb, NULL, DB_STAT_NAME, NULL, DB_BTREE,
+			      DB_CREATE, 0644);
+		if (rc)
+		    syslog(LOG_ERR, "BDB-%d: db->open %s: %s",
+			   rc, DB_STAT_NAME, db_strerror(rc));
+	    }
+	}
     }
     return rc;
 }
@@ -171,6 +201,12 @@ cleanup()
 	free(dbkey.data);
     if (policy_request)
 	free(policy_request);
+    if (statsdb) {
+	rc = statsdb->close(statsdb, 0);
+	statsdb = 0;
+	if (rc)
+	    syslog(LOG_ERR, "BDB-%d: statsdb close: %s", rc, db_strerror(rc));
+    }
     if (db) {
 	rc = db->close(db, 0);
 	db = 0;
@@ -189,7 +225,8 @@ cleanup()
 	else {
 	    rc = dbenv->remove(dbenv, db_home, 0);
 	    if (rc)
-		syslog(LOG_WARNING, "db_env->remove failed", rc);
+		syslog(LOG_WARNING, "BDB-%d: db_env->remove failed: %s",
+		       rc, db_strerror(rc));
 	}
     }
     if (debug_me)
@@ -377,54 +414,140 @@ put_grey_data()
 	log_db_error("put", rc);
 }
 
-static const char *
+static void
+stats_expire()
+{
+    int rc;
+    DBT key;
+    DBT data;
+    memset(&key, 0, sizeof key);
+    memset(&data, 0, sizeof data);
+    if (triplet_data.pass_count) {
+	;
+    }
+    else {
+	unsigned long count;
+	key.data = STAT_BLOCKED;
+	key.size = strlen(STAT_BLOCKED);
+	data.data = &count;
+	data.ulen = sizeof(count);
+	data.flags = DB_DBT_USERMEM;
+	rc = statsdb->get(statsdb, NULL, &key, &data, 0);
+	if (rc == DB_NOTFOUND)
+	    count = 1;
+	else if (rc) {
+	    syslog(LOG_ERR, "BDB-%d: stats_expire block get: %s",
+		   rc, db_strerror(rc));
+	    return;
+	}
+	rc = statsdb->put(statsdb, NULL, &key, &data, 0);
+	if (rc)
+	    syslog(LOG_ERR, "BDB-%d: stats_expire block put: %s",
+		   rc, db_strerror(rc));
+    }
+}
+
+static int
+triplet_expired_p(const struct triplet_data *data)
+{
+    unsigned int age = triplet_data.access_time - triplet_data.create_time;
+    return triplet_data.pass_count == 0 && age > AUTO_RECORD_LIFE_SECS
+	|| triplet_data.pass_count != 0 && age > UPDATE_RECORD_LIFE_SECS;
+}
+
+static void
 process_smtp_rcpt()
 {
-    const char *action = "DUNNO";
+    unsigned long delay;
     get_grey_data();
-    /* Expire record created from a new mail (ip, from, to) */
-    if (triplet_data.pass_count == 0
-	&& triplet_data.access_time - triplet_data.create_time > AUTO_RECORD_LIFE_SECS) {
+    /* Expire records */
+    if (triplet_expired_p(&triplet_data)) {
 	if (debug_me)
 	    syslog(LOG_DEBUG, "expired record");
+	stats_expire();
 	triplet_data.create_time = triplet_data.access_time;
     }
+    delay = triplet_data.access_time - triplet_data.create_time;
     /* Block inbound mail that is from a previously unknown (ip, from, to) triplet */
-    if (triplet_data.access_time - triplet_data.create_time < DELAY_MAIL_SECS) {
+    if (delay < DELAY_MAIL_SECS) {
+	delay = DELAY_MAIL_SECS - delay;
 	triplet_data.block_count++;
-	action = "DEFER_IF_PERMIT Greylisted, try again latter.  See http://projects.puremagic.com/greylisting/ for more information.";
+	printf("action=DEFER_IF_PERMIT Greylisted by " PACKAGE_STRING ", "
+	       "try again in %lu second%s."
+	       "  See http://projects.puremagic.com/greylisting/ for more information.\n",
+	       delay, delay == 1 ? "" : "s");
     }
+    else if (triplet_data.pass_count++)
+	puts("action=DUNNO");
     else
-	triplet_data.pass_count++;
+	printf("action=PREPEND X-Greyfix: Greylisted by " PACKAGE_STRING
+	       " for %lu second%s\n",
+	       delay, delay == 1 ? "" : "s");
     put_grey_data();
-    return action;
+}
+
+static void
+signal_handler(int signal)
+{
+    cleanup();
+    syslog(LOG_NOTICE, "Received signal %d", signal);
+    kill(getpid(), signal);
+    exit(-1);
 }
 
 int
 main(int argc, const char **argv)
 {
-    const char **parg;
+    unsigned int i;
     int rc;
     progname = strrchr(argv[0], '/');
     if (progname)
 	progname++;
     else
 	progname = argv[0];
-    openlog(progname, LOG_PID, LOG_DAEMON);
-    for (parg = argv; *parg; parg++) {
-	if (strcmp("-d", *parg) == 0)
+    openlog(progname, LOG_PID, LOG_MAIL);
+    for (i = 1; i < argc; i++) {
+	if (strcmp("-d", argv[i]) == 0)
 	    debug_me = 1;
-	else if (strcmp("-v", *parg) == 0)
+	else if (strcmp("-v", argv[i]) == 0)
 	    opt_verbose++;
-	else if (strcmp("-h", *parg) == 0) {
-	    parg++;
-	    if (*parg == 0)
+	else if (strcmp("-h", argv[i]) == 0) {
+	    if (++i >= argc)
 		fatal("Missing argument to -h");
-	    db_home = *parg;
+	    db_home = argv[i];
+	}
+	else {
+	    fprintf(stderr, "Unknown option \"%s\"\n", argv[i]);
+	    exit(EXIT_FAILURE);
 	}
     }
     if (opt_verbose)
         syslog(LOG_NOTICE, "daemon started");
+#ifdef SIGHUP
+    signal(SIGHUP, signal_handler);
+#endif
+#ifdef SIGINT
+    signal(SIGINT, signal_handler);
+#endif
+#ifdef SIGQUIT
+    signal(SIGQUIT, signal_handler);
+#endif
+#ifdef SIGILL
+    signal(SIGILL, signal_handler);
+#endif
+#ifdef SIGABRT
+    signal(SIGABRT, signal_handler);
+#endif
+#ifdef SIGSEGV
+    signal(SIGSEGV, signal_handler);
+#endif
+#ifdef SIGALRM
+    signal(SIGALRM, signal_handler);
+#endif
+#ifdef SIGTERM
+    signal(SIGTERM, signal_handler);
+#endif
+
     rc = initialize();
     if (rc) {
 	errno = rc;
@@ -432,7 +555,6 @@ main(int argc, const char **argv)
     }
     while (read_policy_request(0)) {
 	const char *protocol = 0, *state = 0, *ip, *from, *to;
-	const char *action = "DUNNO";
 	if ((protocol = find_attribute("protocol_name"))
 	    && (state = find_attribute("protocol_state"))
 	    && (prefixp(protocol, str_smtp) || prefixp(protocol, str_esmtp))
@@ -441,30 +563,32 @@ main(int argc, const char **argv)
 	    && (from = find_attribute("sender"))
 	    && (to = find_attribute("recipient"))) {
 	    build_triplet_key(ip, from, to);
-	    action = process_smtp_rcpt();
+	    process_smtp_rcpt();
 	}
-	else if (debug_me) {
-	    char *p = 0, *s = 0;
-	    if (protocol) {
-		p = strchr(protocol, '\n');
-		*p = 0;
+	else {
+	    puts("action=DUNNO");
+	    if (debug_me) {
+		char *p = 0, *s = 0;
+		if (protocol) {
+		    p = strchr(protocol, '\n');
+		    *p = 0;
+		}
+		if (state) {
+		    s = strchr(state, '\n');
+		    *s = 0;
+		}
+		syslog(LOG_DEBUG,
+		       "Ignoring protocol %s state %s",
+		       protocol ? protocol : "(not defined)",
+		       state ? state : "(not defined)");
+		if (p)
+		    *p = '\n';
+		if (s)
+		    *s = '\n';
 	    }
-	    if (state) {
-		s = strchr(state, '\n');
-		*s = 0;
-	    }
-	    syslog(LOG_DEBUG,
-		   "Ignoring protocol %s state %s",
-		   protocol ? protocol : "(not defined)",
-		   state ? state : "(not defined)");
-	    if (p)
-		*p = '\n';
-	    if (s)
-		*s = '\n';
 	}
-	safe_writes(1, "action=");
-	safe_writes(1, action);
-	safe_writes(1, "\n\n");
+	putchar('\n');
+	fflush(stdout);
     }
     cleanup();
     if (opt_verbose)
